# GAscripts

This directory gathers all the scripts created to generate a personal version of a general Genetic Algorithm (GA) and several tests.

All the files are modules with functions to be used as working methods, I didn't create a main file with a main process.

There are mostly three kind of code files:

- Files with **GA code**, implementing the genetic algorithms and the functions it requires.

- Files with functions for **analysis**, they have suitable functions to analyze some features of the outputs generated by the previous files.

- Files with **plotting** functions, containing functions to generate figures and short videos.

## Independent files

There are two files independent from any directory:

- [myga]() - Containing `GeneticOptimization`, a method implementing the most general version of a genetic algorithm I produced.
Any new problem can be adapted to be solved by this procedure.

- [plotTools]() - It is a recurrent file containing some functions to represented results graphically.

There are many directories that I used to try different examples and to study deeply the properties of Genetic Algorithms.

## Exercises

This directory contains two simple examples that easily check if the produced GA procedure behaves as expected.

There are two **GA code** files:

- [MaxInteger](Exercises/MaxInteger) - Which implements a GA that looks for the chromosome representing the highest integer in binary representation.

- [MaxNumberOfOnes](Exercises/MaxNumberOfOnes) - Which implements a GA that looks for the chromosome with the maximum number of ones.

This directory also contains some figures that shows some tests that were implemented to see the influence of some parameters.
The three figures show the number of steps needed to find the best chromosome in the `MaxNumberOfOnes` procedure.
Different values for `pc` were tested (horizontal axis) and the number of steps needed counted (vertical axis).
A detailed value sweep was performed for the range 0.5 - 1.0.

## EgFunction

This directory contains the GAs that look for an optimum point in a given function.

- [CurveMax](EgFunction/CurveMax) - Code file that implements a GA that looks for an optimum point in a curve.

- [SurfaceMax](EgFunction/SurfaceMax) - Code file that implements a GA that looks for an optimum point in a surface, this was used to illustrate genetic algorithms in the project memory. 

The remaining files, [plotTools](EgFunction/plotTools) and [SurfaceMax_figuregen](EgFunction/SurfaceMax_figuregen), implement functions to graphically represent the data.

The [data](EgFunction/data) directory contains the output data from the GA with different choice of parameters.
That data was used to generate the images in [Images and animations](EgFunction/Images and animations),
 those images were used in the project memory.
 
 ## PrisonersDilemma
 
 The Prisoners Dilemma is a canonical game example in game theory.
 In a nutshell, two players are forced to decide whether to cooperate or betray each other.
 Both can get more if they are the only one betraying, but if both do so they end up in a wost scenario than if they had cooperated.
 
 In the project memory we used this game to show how genetic algorithms can be used to evolve strategies to play this game.
 Decisions are taken based on the previous decisions of each players, that help the strategies infer if they are trustworthy.
 The example is based on a exercise proposed in
  [M. Melanie, *An introduction to genetic algorithms*. Cambridge, MA, USA: MIT Press, 1996](https://mitpress.mit.edu/books/introduction-genetic-algorithms).
 
 - [PrisonersDilemma](PrisonersDilemma/PrisonersDilemma) - The main file implementing the GA that evolves the desired strategies.
 
 - [PD_behavior_analysis](PrisonersDilemma/PD_behavior_analysis) - Implements some functions to analyze the outputs produced by the previous code.
 
 - [PrisonersDilemma_figuregen](PrisonersDilemma/PrisonersDilemma_figuregen) - Module used to generate graphical representations of the results.
 
 The file [idle_PrisonersDilemmaAnalysis](PrisonersDilemma/idle_PrisonersDilemmaAnalysis)
 is a non-runnable file containing idle code used to get some information with the functions from [PD_behavior_analysis](PrisonersDilemma/PD_behavior_analysis).
 In particular a correlation was found between the fitness of a final population and the tendency to cooperate with strategies that have previously cooperated (cor=+0.5),
 and a negative correlation between the fitness of a final population and its tendency to cooperate with other strategy that had previously betrayed (cor=-0.2).
 This shows a tendency towards cooperate-cooperate.
 
 The Figures directory contains figures generated from different trials, corresponding to the outputs stored in the other directories.
 
 ## CurveFitting
 
 This directory contains a single file [CurveFit](CurveFitting/CurveFit.py) implementing a genetic algorithm that looks for the best parameters to fit a data set with a given parametrized function.
 
 ## SpiderWeb
 
 In this example the building of a web is considered. A web is defined as an instance of the class `Web` contained in [Web](SpiderWeb/Web),
 it is defined by a `Holder` and a set of points describing the start and end of each thread.
 The thread points are encoded in such a way that no hanging thread is allowed, that representation is the one that is used for mutation and crossover.
 The method `flyhaunt` performs a random sample of points with a given radius and measures how many of this points are cached by the web.
 That is used as the fitness criteria in `main`, where the number of threads is limited and a simple selection-crossover procedure is performed.
 An output example is shown in [web_output](SpiderWeb/web_output.png) and [web_fitness_inpop](SpiderWeb/web_fitness_inpop.png).
